from __future__ import annotations

from typing_extensions import override

from great_expectations_cloud.agent.actions.agent_action import (
    ActionResult,
    AgentAction,
)
from great_expectations_cloud.agent.event_handler import register_event_action
from great_expectations_cloud.agent.models import (
    CreatedResource,
    RunCheckpointEvent,
)


class RunCheckpointAction(AgentAction[RunCheckpointEvent]):
    # TODO: New actions need to be created that are compatible with GX v1 and registered for v1.
    #  This action is registered for v0, see register_event_action()

    @override
    def run(self, event: RunCheckpointEvent, id: str) -> ActionResult:
        # TODO: move connection testing into OSS; there isn't really a reason it can't be done there
        for datasource_name, data_asset_names in event.datasource_names_to_asset_names.items():
            datasource = self._context.get_datasource(datasource_name)
            datasource.test_connection(test_assets=False)  # raises `TestConnectionError` on failure
            for data_asset_name in (
                data_asset_names
            ):  # only test connection for assets that are validated in checkpoint
                asset = datasource.get_asset(data_asset_name)
                asset.test_connection()  # raises `TestConnectionError` on failure
        checkpoint_run_result = self._context.run_checkpoint(
            ge_cloud_id=event.checkpoint_id,
            batch_request={"options": event.splitter_options} if event.splitter_options else None,
        )

        validation_results = checkpoint_run_result.run_results
        created_resources = []
        for key in validation_results.keys():
            created_resource = CreatedResource(
                resource_id=validation_results[key]["actions_results"]["store_validation_result"][
                    "id"
                ],
                type="SuiteValidationResult",
            )
            created_resources.append(created_resource)

        return ActionResult(
            id=id,
            type=event.type,
            created_resources=created_resources,
        )


register_event_action("0", RunCheckpointEvent, RunCheckpointAction)



from __future__ import annotations

import enum
from typing import Final


class HeaderName(str, enum.Enum):
    USER_AGENT = "User-Agent"
    AGENT_JOB_ID = "Agent-Job-Id"


USER_AGENT_HEADER: Final = "gx-agent"

__all__ = ["USER_AGENT_HEADER"]



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, Sequence, TypeVar

from great_expectations.compatibility.pydantic import BaseModel

from great_expectations_cloud.agent.models import CreatedResource, Event

if TYPE_CHECKING:
    from great_expectations.data_context import CloudDataContext


class ActionResult(BaseModel):  # type: ignore[misc] # BaseSettings is has Any type
    id: str
    type: str
    created_resources: Sequence[CreatedResource]


_EventT = TypeVar("_EventT", bound=Event)


class AgentAction(Generic[_EventT]):
    def __init__(self, context: CloudDataContext):
        self._context = context

    @abstractmethod
    def run(self, event: _EventT, id: str) -> ActionResult: ...



from __future__ import annotations

import asyncio
import ssl
from asyncio import AbstractEventLoop
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Callable, Protocol

import pika
from pika.adapters.asyncio_connection import AsyncioConnection

if TYPE_CHECKING:
    from pika.channel import Channel
    from pika.spec import Basic, BasicProperties


@dataclass(frozen=True)
class OnMessagePayload:
    correlation_id: str
    delivery_tag: int
    body: bytes


class OnMessageFn(Protocol):
    """
    Callback invoked when a message is received.
    Accepts a single argument, a payload object and returns None.
    """

    def __call__(self, payload: OnMessagePayload) -> None: ...


class AsyncRabbitMQClient:
    """Configuration for a particular AMQP client library."""

    def __init__(self, url: str):
        self._parameters = self._build_client_parameters(url=url)
        self.should_reconnect = False
        self.was_consuming = False
        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._consuming = False

    def run(self, queue: str, on_message: OnMessageFn) -> None:
        """Run an async connection to RabbitMQ.

        Args:
            queue: string representing queue to subscribe to
            on_message: callback which will be invoked when a message is received.
        """
        # When Pika receives an incoming message, our method _callback_handler will
        # be invoked. Here we add the user-provided callback to its signature:
        on_message_callback = partial(self._callback_handler, on_message=on_message)
        # We pass the curried on_message_callback and queue name to _on_connection_open,
        # which Pika will invoke after RabbitMQ establishes our connection.
        on_connection_open_callback = partial(
            self._on_connection_open, queue=queue, on_message=on_message_callback
        )
        connection = AsyncioConnection(
            parameters=self._parameters,
            on_open_callback=on_connection_open_callback,
            on_open_error_callback=self._on_connection_open_error,
            on_close_callback=self._on_connection_closed,
        )
        self._connection = connection
        connection.ioloop.run_forever()

    def stop(self) -> None:
        """Close the connection to RabbitMQ."""
        if self._connection is None:
            return
        if not self._closing:
            self._closing = True
            if self._consuming:
                self._stop_consuming()
                self._connection.ioloop.run_forever()
            else:
                self._connection.ioloop.stop()

    def reset(self) -> None:
        """Reset client to allow a restart."""
        self.should_reconnect = False
        self.was_consuming = False

        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._consuming = False

    def nack(self, delivery_tag: int, requeue: bool) -> None:
        """Nack a message, and indicate if it should be requeued.

        Note that this method is not threadsafe, and must be invoked in the main thread.
        """
        if self._channel is not None:
            self._channel.basic_nack(delivery_tag=delivery_tag, requeue=requeue)

    def get_threadsafe_ack_callback(self, delivery_tag: int) -> Callable[[], None]:
        """Get a callback to ack a message from any thread."""
        if self._channel is None:
            # if the channel is gone, we can't ack
            return lambda: None
        loop = asyncio.get_event_loop()
        ack = partial(
            self._ack_threadsafe,
            channel=self._channel,
            loop=loop,
            delivery_tag=delivery_tag,
        )
        return ack

    def get_threadsafe_nack_callback(
        self, delivery_tag: int, requeue: bool = False
    ) -> Callable[[], None]:
        """Get a callback to nack a message from any thread."""
        if self._channel is None:
            # if the channel is gone, we can't nack
            return lambda: None
        loop = asyncio.get_event_loop()
        nack = partial(
            self._nack_threadsafe,
            channel=self._channel,
            loop=loop,
            delivery_tag=delivery_tag,
            requeue=requeue,
        )
        return nack

    def _ack_threadsafe(self, channel: Channel, delivery_tag: int, loop: AbstractEventLoop) -> None:
        """Ack a message in a threadsafe manner."""
        if channel.is_closed is not True:
            ack = partial(channel.basic_ack, delivery_tag=delivery_tag)
            loop.call_soon_threadsafe(callback=ack)

    def _nack_threadsafe(
        self,
        channel: Channel,
        delivery_tag: int,
        loop: AbstractEventLoop,
        requeue: bool,
    ) -> None:
        """Nack a message in a threadsafe manner."""
        if channel.is_closed is not True:
            nack = partial(channel.basic_nack, delivery_tag=delivery_tag, requeue=requeue)
            loop.call_soon_threadsafe(callback=nack)

    def _callback_handler(  # noqa: PLR0913
        self,
        channel: Channel,
        method_frame: Basic.Deliver,
        header_frame: BasicProperties,
        body: bytes,
        on_message: OnMessageFn,
    ) -> None:
        """Called by Pika when a message is received."""
        # param on_message is provided by the caller as an argument to AsyncRabbitMQClient.run
        correlation_id = header_frame.correlation_id
        delivery_tag = method_frame.delivery_tag
        payload = OnMessagePayload(
            correlation_id=correlation_id, delivery_tag=delivery_tag, body=body
        )
        return on_message(payload)

    def _start_consuming(self, queue: str, on_message: OnMessageFn, channel: Channel) -> None:
        """Consume from a channel with the on_message callback."""
        channel.add_on_cancel_callback(self._on_consumer_canceled)
        self._consumer_tag = channel.basic_consume(queue=queue, on_message_callback=on_message)

    def _on_consumer_canceled(self, method_frame: Basic.Cancel) -> None:
        """Callback invoked when the broker cancels the client's connection."""
        if self._channel is not None:
            self._channel.close()

    def _reconnect(self) -> None:
        """Prepare the client to reconnect."""
        self.should_reconnect = True
        self.stop()

    def _stop_consuming(self) -> None:
        """Cancel the channel, if it exists."""
        if self._channel is not None:
            self._channel.basic_cancel(self._consumer_tag, callback=self._on_cancel_ok)

    def _on_cancel_ok(self, method_frame: Basic.CancelOk) -> None:
        """Callback invoked after broker confirms cancel."""
        self._consuming = False
        if self._channel is not None:
            self._channel.close()

    def _on_connection_open(
        self, connection: AsyncioConnection, queue: str, on_message: OnMessageFn
    ) -> None:
        """Callback invoked after the broker opens the connection."""
        on_channel_open = partial(self._on_channel_open, queue=queue, on_message=on_message)
        connection.channel(on_open_callback=on_channel_open)

    def _on_connection_open_error(self, connection: AsyncioConnection, reason: str) -> None:
        """Callback invoked when there is an error while opening connection."""
        self._reconnect()

    def _on_connection_closed(self, connection: AsyncioConnection, reason: str) -> None:
        """Callback invoked after the broker closes the connection"""
        self._channel = None
        if self._closing:
            connection.ioloop.stop()
        else:
            self._reconnect()

    def _close_connection(self) -> None:
        """Close the connection to the broker."""
        self._consuming = False
        if self._connection is None or self._connection.is_closing or self._connection.is_closed:
            pass
        else:
            self._connection.close()

    def _on_channel_open(self, channel: Channel, queue: str, on_message: OnMessageFn) -> None:
        """Callback invoked after the broker opens the channel."""
        self._channel = channel
        channel.add_on_close_callback(self._on_channel_closed)
        self._start_consuming(queue=queue, on_message=on_message, channel=channel)

    def _on_channel_closed(self, channel: Channel, reason: str) -> None:
        """Callback invoked after the broker closes the channel."""
        self._close_connection()

    def _build_client_parameters(self, url: str) -> pika.URLParameters:
        """Configure parameters used to connect to the broker."""
        parameters = pika.URLParameters(url)
        # only enable SSL if connection string calls for it
        if url.startswith("amqps://"):
            # SSL Context for TLS configuration of Amazon MQ for RabbitMQ
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_default_certs()
            ssl_context.set_ciphers("ECDHE+AESGCM:!ECDSA")
            parameters.ssl_options = pika.SSLOptions(context=ssl_context)
        return parameters


class ClientError(Exception): ...



from __future__ import annotations

import logging
from collections import defaultdict
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any, Final

import great_expectations as gx
from great_expectations.compatibility import pydantic
from packaging.version import Version
from packaging.version import parse as parse_version

from great_expectations_cloud.agent.actions.unknown import UnknownEventAction
from great_expectations_cloud.agent.models import (
    Event,
    UnknownEvent,
)

if TYPE_CHECKING:
    from great_expectations.data_context import CloudDataContext

    from great_expectations_cloud.agent.actions.agent_action import ActionResult, AgentAction


LOGGER: Final[logging.Logger] = logging.getLogger(__name__)


# Map of event types to actions, specific to the GX Core major version.
# Built via registering using register_event_action(). For example:
# _EVENT_ACTION_MAP = {
#     "0": {
#         "EventName": AgentActionClass,
#         "RunCheckpointEvent": RunCheckpointAction,
#     },
#     "1": {
#         # No events implemented yet for GX Core v1
#         # Suggested naming convention, add version suffix "V1" e.g.: RunCheckpointEventV1
#     },
# }
_EVENT_ACTION_MAP: dict[str, dict[str, type[AgentAction[Any]]]] = defaultdict(dict)


def register_event_action(
    version: str, event_type: type[Event], action_class: type[AgentAction[Any]]
) -> None:
    """Register an event type to an action class."""
    if version in _EVENT_ACTION_MAP and event_type.__name__ in _EVENT_ACTION_MAP[version]:
        raise EventAlreadyRegisteredError(event_type_name=event_type.__name__, version=version)
    event_type_str = event_type.__name__
    _EVENT_ACTION_MAP[version][event_type_str] = action_class
    LOGGER.debug(
        f"Registered event action: {event_type_str} -> {action_class.__name__} (version {version})"
    )


class EventHandler:
    """
    Core business logic mapping events to actions.
    """

    def __init__(self, context: CloudDataContext) -> None:
        self._context = context

    def get_event_action(self, event: Event) -> AgentAction[Any]:
        """Get the action that should be run for the given event."""
        action_map = _EVENT_ACTION_MAP.get(_GX_MAJOR_VERSION)
        if action_map is None:
            raise NoVersionImplementationError(version=_GX_MAJOR_VERSION)
        action_class = action_map.get(_get_event_name(event))
        if action_class is None:
            action_class = UnknownEventAction
        return action_class(context=self._context)

    def handle_event(self, event: Event, id: str) -> ActionResult:
        """Transform an Event into an ActionResult."""
        action = self.get_event_action(event=event)
        LOGGER.info(f"Handling event: {event.type} -> {action.__class__.__name__}")
        action_result = action.run(event=event, id=id)
        return action_result

    @classmethod
    def parse_event_from(cls, msg_body: bytes) -> Event:
        try:
            event: Event = pydantic.parse_raw_as(Event, msg_body)
        except (pydantic.ValidationError, JSONDecodeError):
            # Log as bytes
            LOGGER.exception("Unable to parse event type", extra={"msg_body": f"{msg_body!r}"})
            return UnknownEvent()

        return event


class EventError(Exception): ...


class UnknownEventError(EventError):
    def __init__(self, event_name: str):
        super().__init__(f'Unknown message received: "{event_name}" - cannot process.')


class NoVersionImplementationError(EventError):
    def __init__(self, version: str | Version):
        super().__init__(f"No event action map implemented for GX Core major version {version}.")


class InvalidVersionError(EventError):
    def __init__(self, version: str | Version):
        super().__init__(f"Invalid version: {version}")


class EventAlreadyRegisteredError(EventError):
    def __init__(self, event_type_name: str, version: str | Version):
        super().__init__(f"Event type {event_type_name} already registered for version {version}.")


def _get_major_version(version: str) -> str:
    """Get major version as a string from version as a string. For example, "0.18.0" -> "0"."""
    parsed: Version = parse_version(version)
    if not isinstance(parsed, Version):
        raise InvalidVersionError(version)
    return str(parsed.major)


version = gx.__version__  # type: ignore[attr-defined] # TODO: fix this
_GX_MAJOR_VERSION = _get_major_version(str(version))


def _get_event_name(event: Event) -> str:
    try:
        return str(event.__name__)
    except AttributeError:
        return str(event.__class__.__name__)



from __future__ import annotations

import argparse
import dataclasses as dc
import json
import logging
import pathlib
import sys
from typing import Any

from great_expectations_cloud.logging.logging_cfg import LogLevel, LogSettings, configure_logger

LOGGER = logging.getLogger(__name__)


@dc.dataclass(frozen=True)
class Arguments:
    log_level: LogLevel
    skip_log_file: bool
    json_log: bool
    log_cfg_file: pathlib.Path | None
    version: bool
    custom_log_tags: str


def _parse_args() -> Arguments:
    """
    Parse arguments from the command line and return them as a type aware
    `Arguments` dataclass.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log-level",
        help="Level of logging to use. Defaults to WARNING.",
        default="WARNING",
        type=LogLevel,
    )
    parser.add_argument(
        "--skip-log-file",
        help="Skip writing debug logs to a file. Defaults to False. Does not affect logging to stdout/stderr.",
        default=False,
        type=pathlib.Path,
    )
    parser.add_argument(
        "--json-log",
        help="Output logs in JSON format. Defaults to False.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--log-cfg-file",
        help="Path to a logging configuration file in JSON format. Supersedes --log-level and --skip-log-file.",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--custom-log-tags",
        help="Additional tags for logs in form of key-value pairs",
        type=str,
        default="{}",
    )
    parser.add_argument("--version", help="Show the gx agent version.", action="store_true")
    args = parser.parse_args()
    return Arguments(
        log_level=args.log_level,
        skip_log_file=args.skip_log_file,
        log_cfg_file=args.log_cfg_file,
        version=args.version,
        json_log=args.json_log,
        custom_log_tags=args.custom_log_tags,
    )


def main() -> None:
    # lazy imports ensure our cli is fast and responsive
    args: Arguments = _parse_args()
    custom_tags: dict[str, Any] = {}
    try:
        custom_tags = json.loads(args.custom_log_tags)
    except json.JSONDecodeError as e:
        print(f"Failed to parse custom tags {args.custom_log_tags} due to {e}")
        sys.exit(1)

    configure_logger(
        LogSettings(
            log_level=args.log_level,
            skip_log_file=args.skip_log_file,
            log_cfg_file=args.log_cfg_file,
            json_log=args.json_log,
            custom_tags=custom_tags,
        )
    )

    if args.version:
        from great_expectations_cloud.agent import get_version

        print(f"GX Agent version: {get_version()}")
        return

    from great_expectations_cloud.agent import run_agent

    run_agent()


if __name__ == "__main__":
    main()
